/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* Copyright 2013 - 2024, nymea GmbH
* Contact: contact@nymea.io
*
* This file is part of nymea.
* This project including source code and documentation is protected by
* copyright law, and remains the property of nymea GmbH. All rights, including
* reproduction, publication, editing and translation, are reserved. The use of
* this project is subject to the terms of a license agreement to be concluded
* with nymea GmbH in accordance with the terms of use of nymea GmbH, available
* under https://nymea.io/license
*
* GNU Lesser General Public License Usage
* Alternatively, this project may be redistributed and/or modified under the
* terms of the GNU Lesser General Public License as published by the Free
* Software Foundation; version 3. This project is distributed in the hope that
* it will be useful, but WITHOUT ANY WARRANTY; without even the implied
* warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License
* along with this project. If not, see <https://www.gnu.org/licenses/>.
*
* For any further details and any questions please contact us under
* contact@nymea.io or see our FAQ/Licensing Information on
* https://nymea.io/license/faq
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef QUEUEDMODBUSREPLY_H
#define QUEUEDMODBUSREPLY_H

#include <QObject>
#include <QModbusReply>
#include <QModbusDataUnit>

class QueuedModbusReply : public QObject
{
    Q_OBJECT
public:
    enum RequestType {
        RequestTypeRead,
        RequestTypeWrite
    };
    Q_ENUM(RequestType)

    explicit QueuedModbusReply(QObject *parent = nullptr);
    explicit QueuedModbusReply(RequestType requestType, QModbusDataUnit dataUnit, QObject *parent = nullptr);

    ~QueuedModbusReply();

    RequestType requestType() const;
    void setRequestType(RequestType requestType);

    QModbusDataUnit dataUnit() const;
    void setQMOdbusDataUnit(const QModbusDataUnit &dataUnit);

    // Available once the request will be sent to the modbus slave
    QModbusReply *reply() const;
    void setReply(QModbusReply *reply);

    QModbusDevice::Error error() const;
    QString errorString() const;

signals:
    void finished();
    void errorOccurred(QModbusDevice::Error error);

private:
    RequestType m_requestType = RequestTypeRead;
    QModbusDataUnit m_dataUnit;
    QModbusReply *m_reply = nullptr;

};

#endif // QUEUEDMODBUSREPLY_H
